
import numpy as numpy
import math
from . import Tools
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from .Sample import Sample
from .SampleList import SampleList

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2020 Do Kester


class SampleMovie( object ):
    """
    SampleMovie produces a movie (mp4) from a SampleList

    This class is provided as an example to vary upon.

    ===========
    MovieWriter
    ===========

    This example uses a MovieWriter directly to grab individual frames and write
    them to a file. This avoids any event loop integration, but has the advantage
    of working with even the Agg backend. This is not recommended for use in an
    interactive setting.

    """

    def __init__( self, samplelist, filename="samplemovie.mp4", problem=None, kpar=[0,1] ) :
        """
        Constructor.

        The constructor produces the movie.

        Parameters
        ----------
        samplelist : SampleList
            to make the movie from
        filename : str
            name of the mp4 movie
        problem : Problem
            the problem that produced the samplelist
        kpar : list of 2 ints
            indices of the parameters to plot
        """

        bakend = matplotlib.get_backend()       ## save present backend to restore

        matplotlib.use("Agg")					## need to use this one to make a movie

        FFMpegWriter = animation.writers[ 'ffmpeg' ]
        metadata = dict( title='Sample Movie', artist='BayesicFitting',
                comment='Movie support!' )
        writer = FFMpegWriter( fps=30, metadata=metadata )

        fig = plt.figure()

        if problem is not None :
            ld, lm, l1, l2, l4 = plt.plot( [], [], 'b*', [], [], 'g-',
                                           [], [], 'y.', [], [], 'r.', [], [], 'k+' )
            xd = problem.xdata
            yd = problem.ydata
            xm = numpy.linspace( numpy.min( xd ), numpy.max( xd ), 10 * len( xd ), dtype=float )
            # xd -= 5.0
            yd -= 9.0
            xn = xm # - 5.0
        else :
            l1, l2, l4 = plt.plot( [], [], 'y.', [], [], 'r.', [], [], 'k+' )

        smpl = samplelist[0]
        k0 = kpar[0]
        k1 = kpar[1]
        pr0 = smpl.model.priors[k0]
        plt.xlim( pr0.lowLimit, pr0.highLimit )
        plt.xlabel( smpl.model.getParameterName( k0 ) )
        pr1 = smpl.model.priors[k1]
        plt.ylim( pr1.lowLimit, pr1.highLimit )
        plt.ylabel( smpl.model.getParameterName( k1 ) )

        mx = smpl.logW
        ns = len( samplelist )
        LOG100 = math.log( 100 )
        with writer.saving( fig, filename, 100 ):
            for i in range( ns ):
                s = samplelist[i]
                if problem is not None :
                    ym = s.model.result( xm, s.parameters ) - 9.0
                    ld.set_data( xd, yd )
                    lm.set_data( xn, ym )
                x1 = []
                y1 = []
                x4 = []
                y4 = []
                if samplelist[i].logW > mx :
                    mx = samplelist[i].logW
                if samplelist[i].logW < ( mx - LOG100 ) :
                    return
                for sample in samplelist[:i] :
                    if sample.logW < ( mx - LOG100 ) :
                        x1 = x1 + [sample.allpars[k0]]
                        y1 = y1 + [sample.allpars[k1]]
                    else :
                        x4 = x4 + [sample.allpars[k0]]
                        y4 = y4 + [sample.allpars[k1]]

                l1.set_data( x1, y1 )
                l4.set_data( x4, y4 )

                x2 = []
                y2 = []

                for sample in samplelist[i:] :
                    if sample.start > i : continue
                    x2 = x2 + [sample.allpars[k0]]
                    y2 = y2 + [sample.allpars[k1]]
                l2.set_data( x2, y2 )

#                print( i, s.id, s.parent, s.start, s.logW, mx, len(x1), len(x2), len(x4) )
                writer.grab_frame( )

        matplotlib.use( bakend )            ## reset to what it was




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt

from .NonLinearModel import NonLinearModel
from .kernels.Kernel import Kernel
from .kernels.Gauss import Gauss

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2022 Do Kester

class Kernel2dModel( NonLinearModel ):
    """
    Two dimensional Kernel Model.

    The Kernel2dModel is defined as

        f( x:p ) = p_0 * K( r )

    where K( r ) is a selectable kernel function and r is the distance to the center.

        r = sqrt( u^2 + v^2 ).

    There are 3 options for u and v

    1. CIRCULAR has 4 parameters
        Circular shape with only one width.
        u = ( x - p_1 ) / p_3
        v = ( x - p_2 ) / p_3
    2. ELLIPTIC has 5 parameters
        Elliptic shape aligned along the axes; 2 widths.
        u = ( x - p_1 ) / p_3
        v = ( x - p_2 ) / p_4
    3. ROTATED has 6 parameters
        Rotated elliptical shape with 2 width and a rotational angle.
        u = ( ( x - p_1 )*cos( p_5 ) - ( y - p_2 )*sin( p_5) ) / p_3
        v = ( ( x - p_1 )*sin( p_5 ) + ( y - p_2 )*cos( p_5) ) / p_4

    The "center" parameters ( 1&2 ) and the "angle" parameter ( 5 ) are initilized as 0.
    The rotational angle is measured counterclockwise from the x-axis.
    The "width" parameters ( 3&4 ) are initialized as 1.0, except for the ROTATED case;
    then they need to be different ( 2.0 and 0.5 resp. ).
    Otherwise the model parameter "angle" is degenerate.
    The "amplitude" parameter is set to 1.0.

    Several kernel functions, K( x ) are defined in the directory fit/kernels.

    Beware: These models are unaware of anything outside their range.

    Author:      Do Kester

    Example
    -------
    >>> model = Kernel2dModel( )                                 # default: circular Gauss
    >>> model.setKernelShape( Lorentz(), 'Elliptic'  )             # elliptic Lorentz model.
    >>> model = Kernel2dModel( shape=3 )                         # rotated Gauss

    Category:    mathematics/Fitting

    """
    SHAPENAMES = ["Circular", "Circular", "Elliptic", "Rotated"]

    CIRCULAR = 1
    ELLIPTIC = 2
    ROTATED = 3
    NPAR = [4,4,5,6]

    def __init__( self, kernel=Gauss(), shape=1, copy=None, **kwargs ):
        """
        Kernel Model.

        Default model: Gauss with Circular shape.

        Parameters
        ----------
        kernel : Kernel
            the kernel to be used
        shape : 1 | 2 | 3 | 'circular' | 'elliptic' | 'rotated'
            int : resp.: circular elliptic, rotated
            str : case insensitive; only the first letter matters.
            shape defaults to 'circular' when misunderstood
        copy : Kernel2dModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """

        shape = self.parseShape( shape )

        super( Kernel2dModel, self ).__init__( self.NPAR[shape], ndim=2, copy=copy, **kwargs )

        if copy is not None :
            self.kernel = copy.kernel
            self.shape = copy.shape
            self.shape2d = copy.shape2d
        else :
            self.setKernelShape( kernel, shape )

    def copy( self ):
        """ Copy method.  """
        return Kernel2dModel( copy=self, shape=self.shape )

    def __setattr__( self, name, value ) :
        if name == "kernel" :
            setatt( self, name, value, type=Kernel )
        elif name == "shape2d" :
            setatt( self, name, value, type=BaseShape2d )
        elif name == "shape" :
            setatt( self, name, value )
        elif name == "npbase" :
            setatt( self, name, value )
        else :
            super( Kernel2dModel, self ).__setattr__( name, value )

    def parseShape( self, shape ) :
        if isinstance( shape, str ) :
            shape = str.find( 'xxcCeErR', shape[0] ) // 2
        if not 1 <= shape <= 3 :
            shape = 1
        return shape


    def setKernelShape( self, kernel, shape ) :
        shape = self.parseShape( shape )
        self.kernel = kernel
        amp = 2.0 / ( self.kernel.integral * math.pi )

        self.shape = shape
        self.npbase = self.NPAR[shape]
        if self.fixed is not None : self.npbase -= len( self.fixed )

        parNames = ["amplitude", "x-center", "y-center"]
        if shape == 1 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 1.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3]
            self.nonZero = [3]
            parNames += ["width"]
            self.shape2d = Circle( self.npbase, self.kernel )
        elif shape == 2 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 1.0, 1.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3,4]
            self.nonZero = [3,4]
            parNames += ["x-width", "y-width"]
            self.shape2d = Ellipse( self.npbase, self.kernel )
        elif shape == 3 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 2.0, 0.5, 0.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3,4]
            self.nonZero = [3,4]
            parNames += ["x-width", "y-width", "angle"]
            self.shape2d = Rotated( self.npbase, self.kernel )
        self.parNames = self.selectNames( parNames )

    def baseName( self ):
        """ Returns a string representation of the model.  """
        return "2-d-" + self.SHAPENAMES[self.shape] + "-" + self.kernel.name( )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return self.shape2d.result( xdata, params )

    def baseDerivative( self, xdata, params ):
        """
        Returns df/dx of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        df = self.shape2d.derivative( xdata, params )
        return numpy.asarray( df ).T

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)
        parlist : array_like
            list of indices of active parameters

        """
        return self.shape2d.partial( xdata, params, parlist=parlist )


    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        if k == 0:
            return self.yUnit
        if k == 5:
            return units.Unit( units.si.rad )
        return self.xUnit[(k-1)%2]

    def isBound( self ):
        return self.kernel.isBound( )

class BaseShape2d( object ):
    def __init__( self, npbase, kernel ) :
        self.npbase = npbase
        self.kernel = kernel

    def reslt( self, xdata, params, kp ):
        x = ( xdata[:,0] - params[1] ) / params[3]
        y = ( xdata[:,1] - params[2] ) / params[kp]
        return params[0] * self.kernel.resultsq( x * x + y * y )

    def deriv( self, xdata, params, kp ) :
        x = ( xdata[:,0] - params[1] ) / params[3]
        y = ( xdata[:,1] - params[2] ) / params[kp]
        r = numpy.sqrt( x * x + y * y )
        drdx = x / r
        drdy = y / r
        df = [params[0] * self.kernel.partial( r ) * drdx / params[3],
              params[0] * self.kernel.partial( r ) * drdy / params[kp]]
        return df


class Circle( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Circle, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        return self.reslt( xdata, params, 3 )

    def derivative( self, xdata, params ):
        return self.deriv( xdata, params, 3 )

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        u = ( xdata[:,0] - params[1] ) / params[3]
        v = ( xdata[:,1] - params[2] ) / params[3]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = -params[0] * self.kernel.partial( r )
        try :
            drdu = numpy.where( r == 0, 1.0, u / r )
            drdv = numpy.where( r == 0, 1.0, v / r )
        except Warning :
            pass

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                  1 : ( lambda: p0dkdr * drdu / params[3] ),
                  2 : ( lambda: p0dkdr * drdv / params[3] ),
                  3 : ( lambda: p0dkdr * r / params[3] ) }

        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def __str__( self ) :
        return "Circle"

class Ellipse( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Ellipse, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        return self.reslt( xdata, params, 4 )

    def derivative( self, xdata, params ):
        return self.deriv( xdata, params, 4 )

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        u = ( xdata[:,0] - params[1] ) / params[3]
        v = ( xdata[:,1] - params[2] ) / params[4]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = -params[0] * self.kernel.partial( r )
        try :
            drdu = numpy.where( r == 0, 1.0, u / r )
            drdv = numpy.where( r == 0, 1.0, v / r )
        except RuntimeWarning :
            pass

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                  1 : ( lambda: p0dkdr * drdu / params[3] ),
                  2 : ( lambda: p0dkdr * drdv / params[4] ),
                  3 : ( lambda: p0dkdr * u * drdu / params[3] ),
                  4 : ( lambda: p0dkdr * v * drdv / params[4] ) }


        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def __str__( self ) :
        return "Ellipse"

class Rotated( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Rotated, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )
        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        return params[0] * self.kernel.resultsq( u * u + v * v )

    def derivative( self, xdata, params ) :
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )

        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        r = numpy.sqrt( u * u + v * v )
        drdu = numpy.where( r == 0, 1.0, u / r )
        drdv = numpy.where( r == 0, 1.0, v / r )
        p0dkdr = params[0] * self.kernel.partial( r )
        df = [ p0dkdr * ( drdu * c / params[3] + drdv * s / params[4] ),
               p0dkdr * ( drdv * c / params[4] - drdu * s / params[3] ) ]

        return df

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )
        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = - params[0] * self.kernel.partial( r )
        try :
            drdu = numpy.where( r == 0, 1.0, u / r )
            drdv = numpy.where( r == 0, 1.0, v / r )
        except RuntimeWarning :
            pass

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                    #  dfdp1 = p0 * dkdr * ( drdu * dudx + drdv * dvdx ) * dxdp1
                  1 : ( lambda: p0dkdr * ( drdu * c / params[3] + drdv * s / params[4] ) ),

                    #  dfdp2 = p0 * dkdr * ( drdu * dudy + drdv * dvdy ) * dydp2
                  2 : ( lambda: p0dkdr * ( -drdu * s / params[3] + drdv * c / params[4] ) ),

                    #  dfdp3 = p0 * dkdr * drdu * dudp3
                  3 : ( lambda: p0dkdr * drdu * u / params[3] ),
                  4 : ( lambda: p0dkdr * drdv * v / params[4] ),

                    #  dfdp5 = p0 * dkdr * ( drdu * ( dudc * dcdp5 + duds * dsdp5 ) +
                    #                        drdv * ( dvdc * dcdp5 + dvds * dsdp5 ) )
                  5 : ( lambda: -p0dkdr * ( drdu * ( -x * s - y * c ) / params[3] +
                                            drdv * ( -y * s + x * c ) / params[4] ) ) }

        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def __str__( self ) :
        return "Rotated"





import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .OrderEngine import OrderEngine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester

class MoveEngine( OrderEngine ):
    """
    The MoveEngine tries to move a selection of the parameters to another spot.

    Input order : [0,1,2,3,4,5,6,7,8,9]

    output order: [4,5,6,7,1,2,3,8,9,0]


    It belongs to the class of generalized travelling salesman problems
    where the parameters of the problem is an ordered list.

    The walker is kept when the logLikelihood > lowLhood

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : OrderEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return MoveEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "MoveEngine" )

    #  *********EXECUTE***************************************************
    def executeOnce( self, kw, lowLhood ) :

        walker = self.walkers[kw]
        problem = walker.problem
        param = walker.allpars

        src = self.rng.randint( problem.npars )
        des = src
        while des == src :
            des = self.rng.randint( problem.npars )

        mx = ( des - src ) if src < des else ( problem.npars - src )
        t = min( self.rng.geometric( 0.2 ), mx )
        while t > 0 :
            if src < des :
                kln = src + t
                ptry = param[kln:des]
                ptry = numpy.append( ptry, param[src:kln] )
                ptry = numpy.append( ptry, param[des:] )
                ptry = numpy.append( ptry, param[:src] )
            else :
                kln = src + t
                ptry = param[kln:]
                ptry = numpy.append( ptry, param[:des] )
                ptry = numpy.append( ptry, param[src:kln] )
                ptry = numpy.append( ptry, param[des:src] )

            Ltry = self.errdis.logLikelihood( problem, ptry )

            if self.verbose > 4 :
                print( "Order    ", src, kln, t, des, fmt( lowLhood ), fmt( Ltry ) )
                print( fmt( param, max=None, format='%3d' ) )
                print( fmt( ptry, max=None, format='%3d' ) )

            if Ltry >= lowLhood:
                self.reportSuccess( )
                self.setWalker( walker.id, problem, ptry, Ltry )

                ## check if better than Lbest in walkers[-1]
                ## self.checkBest( problem, ptry, Ltry )

                return t

            t = self.rng.randint( t )
            self.reportFailed( )


        self.reportReject()

        return t                        # nr of reordered parameters





import numpy as numpy
from . import Tools
from .Tools import setAttribute as setatt
from .SplinesModel import SplinesModel
from .BasicSplinesModel import BasicSplinesModel
from .Dynamic import Dynamic
from .Modifiable import Modifiable

from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2020 - 2020 Do Kester

class SplinesDynamicModel( Modifiable, Dynamic, BasicSplinesModel ):
    """
    BasicSplinesModel that is modifiable (knot locations) and dynamic (in number
    of knots)


    Examples
    --------
    >>> knots = numpy.arange( 17, dtype=float ) * 10    # make equidistant knots from 0 to 160
    >>> csm = SplinesModel( knots=knots, order=2 )
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> csm = SplinesModel( nrknots=17, order=2, min=0, max=160 )    # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> npt = 161                                               # to include both 0 and 160.
    >>> x = numpy.arange( npt, dtype=float )                    # x-values
    >>> csm = SplinesModel( nrknots=17, order=2, xrange=x )     # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18

    Attributes
    ----------
    minKnots : int
        minimum number of knots
    maxDegree : int or None
        maximum number of knots

    Attributes from Modifiable
    --------------------------
        modifiable

    Attributes from Dynamic
    -----------------------
        dynamic, ncomp (=degree+1), deltaNpar, minComp (=minDegree+1), maxComp (=maxDegree+1), growPrior

    Attributes from SplinesModel
    ----------------------------
        knots, order

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Limitations
    -----------
    Dont construct the knots so closely spaced, that there are no datapoints in between.

    """
    def __init__( self, modifiable=True, dynamic=True, growPrior=None, minKnots=2, maxKnots=None,
                        minDistance=0.01, copy=None, **kwargs ):
        """
        Splines on a given set of knots and a given order.

        The number of parameters is ( length( knots ) + order - 1 )

        Parameters
        ----------
        modifiable : bool
            if True allow changement of the knot locations
        dynamic : bool
            if True allow growth and shrinkage of number of knots
        minKnots : int
            minimum number of knots (def=2)
        maxKnots : None or int
            maximum number of Knots
        minDistance : float
            minimum distance between knots, provided as fraction of average knot distance.
            default is ( 0.01 * ( knots[-1] - knots[0] ) / nrknots )
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxDegree is None else UniformPrior
        copy : PolynomialDynamicModel
            model to copy

        Parameters for SplinesModel
        ---------------------------
        knots, order, nrknots, min, max, xrange

        Raises
        ------
        ValueError if not minKnots <= nrknots <= maxKnots

        """
        Modifiable.__init__( self, modifiable=modifiable )
        Dynamic.__init__( self, dynamic=dynamic )

        BasicSplinesModel.__init__( self, copy=copy, **kwargs )

        nrknots = len( self.knots )
        if nrknots < minKnots or ( maxKnots is not None and nrknots > maxKnots ) :
            raise ValueError( "nrknots outside range of [min..max] range" )

        self.deltaNpar = 1

        if copy is None :
            minDistance *= ( self.knots[-1] - self.knots[0] ) / nrknots
            setatt( self, "minDistance", minDistance )
            self.setGrowPrior( growPrior=growPrior, min=minKnots, max=maxKnots,
                           name="Knots" )
        else :
            setatt( self, "knots", copy.knots.copy() )
            setatt( self, "minKnots", copy.minKnots )
            setatt( self, "maxKnots", copy.maxKnots )
            setatt( self, "minDistance", copy.minDistance )
            setatt( self, "growPrior", copy.growPrior.copy() )


    def copy( self ):
        return SplinesDynamicModel( knots=self.knots, modifiable=self.modifiable,
                                    dynamic=self.dynamic, copy=self )

    def __setattr__( self, name, value ) :
        if self.setDynamicAttribute( name, value ) :
            return
        else :
            super( ).__setattr__( name, value )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'ncomp' :
            return len( self.knots )
        if name == 'minComp' :
            return self.minKnots
        if name == 'maxComp' :
            return self.maxKnots

        return super( ).__getattr__( name )

    def baseName( self ):
        """ Returns a string representation of the model. """
        name = "Dyn" if self.dynamic else ""
        name += "Mod" if self.modifiable else ""

        return name + super().baseName()

    def changeNComp( self, dn ) :
        pass                        ## it is the length of knots

    def grow( self, offset=0, rng=None, force=False, **kwargs ):
        """
        Increase the degree by one upto maxComp ( if present ).

        Parameters
        ----------
        offset : int
            index where the params of the Dynamic model start
        rng : random number generator (obligatory)
            to generate a new parameter.
        force : bool
            dont check maxKnots

        Return
        ------
        bool :  succes

        """
        if not force and self.maxKnots is not None and self.knots.size >= self.maxKnots :
            return False

#        print( "SDg p0 ", fmt( self._head.parameters, max=None ) )

        nrknots = len( self.knots )
        npars = self.npbase
        while True :
            newkn = rng.rand() * ( self.knots[-1] - self.knots[0] ) + self.knots[0]
            k = 1
            while self.knots[k] < newkn : k += 1
            if ( newkn - self.knots[k-1] > self.minDistance and
                 self.knots[k] - newkn   > self.minDistance ) :
                break

        self.knots = numpy.insert( self.knots, k, newkn )   ## new set of knots

        location = ( npars - nrknots ) // 2 + k + 1

        dnp = self.deltaNpar
        self.alterParameterSize( dnp, offset, location=location )

        try :
            self.basis = self.makeBasis()
        except :
            print( "Grow: Could not make basis from", fmt( self.knots,max=None ) )
            return False

        head = self._head
        mdlpar = head.parameters
        kl = location + offset

#        print( "SDg p1 ", fmt( mdlpar, max=None ) )

#        print( "grow    ", nrknots, location, offset, k, kl, newkn )

        mdlpar[kl-1] *= 0.7
        value = mdlpar[kl-1]
        mdlpar = self.alterParameters( mdlpar, location, dnp, offset, value=value )

        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( 1 )

        self.changeNComp( 1 )

        return True


    def shrink( self, offset=0, rng=None, **kwargs ):
        """
        Decrease the degree by one downto minComp ( default 1 ).

        Parameters
        ----------
        offset : int
            index where the params of the Dynamic model start
        rng : random number generator
            to generate a new parameter (obligatory)

        Return
        ------
        bool : succes

        """
        nrknots = len( self.knots )
        npars = self.npbase
        if nrknots <= self.minKnots :
            return False

#        print( "SDs p0 ", fmt( self._head.parameters, max=None ) )

        k = rng.randint( nrknots - 2 ) + 1              ## avoid endpoints
        dis = self.knots[k+1] - self.knots[k-1]
        r1 = ( self.knots[k] - self.knots[k-1] ) / dis
        r2 = ( self.knots[k+1] - self.knots[k] ) / dis

        self.knots = numpy.delete( self.knots, k )      ## new set of knots

        location = ( npars - nrknots ) // 2 + k + 1
        dnp = - self.deltaNpar
        kl = offset + location - 1
        head = self._head
        mdlpar = head.parameters

        ## new value is the weighted average of the joined parameters
        value = 1.4 * ( mdlpar[kl] * r1 + mdlpar[kl+1] * r2 )

        if self.getPrior( kl ).isOutOfLimits( value ) :
            value = ( self.getPrior( kl ).stayInLimits( value ) + mdlpar[kl] + mdlpar[kl+1] ) / 3

        if self.getPrior( kl ).isOutOfLimits( value ) :
            print( "New shrink value out of limit" )
            print( "Knots ", k, fmt( r1 ), fmt( r2 ), fmt( dis ), fmt( self.knots, max=None ) )
            print( "Pars  ", fmt( mdlpar, max=None ) )
            print( "Old   ", kl, fmt( mdlpar[kl:kl+2] ), "  New  ", fmt( value ) )
            return False

#        print( "SD  p1 ", npars, nrknots, k, location, kl )

        self.alterParameterSize( dnp, offset, location=location )
        try :
            self.basis = self.makeBasis()
        except :
            print( "Shrk: Could not make basis from", fmt( self.knots,max=None ) )
            return False

        mdlpar = self.alterParameters( mdlpar, location, dnp, offset )
        mdlpar[kl] = value

#        print( "SD  p2 ", fmt( mdlpar, max=None ), kl, value )

        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        self.changeNComp( -1 )

        return True

    def vary( self, rng=None, location=None ) :

        nrknots = len( self.knots )
        if nrknots <= 2 or location == 0 or location == nrknots :
            return False

        kk = location if location is not None else rng.randint( nrknots - 2 ) + 1

        newkn = self.knots[kk]

        while True :
            rn = ( rng.rand() + rng.rand() ) - 1
            newkn += rn * ( ( newkn - self.knots[kk-1] ) if rn < 0 else
                            ( self.knots[kk+1] - newkn ) )

            ## check minimum distance from neighbouring knots
            if ( newkn - self.knots[kk-1] > self.minDistance and
                 self.knots[kk+1] - newkn > self.minDistance ) :
                break

#        print( "SDM   ", fmt( self.knots, max=None ), kk, newkn )

        self.knots[kk] = newkn
        try :
            self.basis = self.makeBasis()
        except :
            print( "Vary: Could not make basis from", fmt( self.knots,max=None ) )
            raise
            return False

        return True

    def varyAlt( self, offset=0, rng=None, **kwargs ) :
        """
        Vary the structure of a Modifiable Model


        Parameters
        ----------
        offset : int
            index where the params of the Modifiable model start
        rng : RNG
            random number generator
        kwargs : keyword arguments
            for specific implementations
        """
        suc = self.grow( offset=offset, rng=rng, force=True )
        suc = suc and self.shrink( offset=offset, rng=rng )
        return suc





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2015 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class EtalonModel( NonLinearModel ):
    """
    Etalon Model.

        f( x:p ) = p_0 / ( 1.0 + p_1 * sin^2( \pi p_2 x + p_3 ) )

    where
        p_0 = amplitude
        p_1 = finesse,
        p_2 = frequency in 1/wavenumber
        p_3 = phase,
    As always x = input; it is in wavenumbers

    The parameters are initialized at {1.0, 1.0, 1.0, 0.0}. It is a non-linear model.

    The finesse should be positive. However, solutions where -1 < p_1 < 0 are equivalent
    to a solution with parameters set as:
        p_0 /= ( 1 + p_1 )
        p_1 /= -( 1 + p_1 )
        p_3 += pi/2             # 90 degree phase shift

    A finesse below -1 causes infinities.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Examples
    --------
    >>> fpm = EtalonModel( )
    >>> print( fpm.npchain )
    4
    >>> pars = [1.0, 30.0, 1.0, 0.0]
    >>> fpm.parameters = pars
    >>> print( fpm( numpy.arange( 101, dtype=float ) ) )     # etalon with 10 periods

    """

    def __init__( self, copy=None, **kwargs ):
        """
        Etalon model.

        Number of parameters is 4.

        Parameters
        ----------
        copy : EtalonModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        param = [1.0, 1.0, 1.0, 0.0]
        names = ["amplitude", "finesse", "frequency", "phase"]
        super( EtalonModel, self ).__init__( 4, copy=copy, params=param,
                    names=names, **kwargs )


    def copy( self ):
        """ Copy method.  """
        return EtalonModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = math.pi * xdata * params[2] + params[3]
        sx = numpy.sin( x )
        return params[0] / ( 1.0 + params[1] * sx * sx )

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = math.pi * xdata * params[2] + params[3]
        sx = numpy.sin( x )
        dd = 1 + params[1] * sx * sx
        dd *= dd
        return - 2 * math.pi * params[0] * params[1] * params[2] * sx * numpy.cos( x ) / dd


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata ), np ) )

        x = math.pi * xdata * params[2] + params[3]
        sx = numpy.sin( x )
        s2 = sx * sx
        dd = 1.0 / ( 1 + params[1] * s2 )
        d2 = dd * dd
        p3 = - 2 * params[0] * params[1] * sx * numpy.cos( x ) * d2

        parts = { 0 : ( lambda: dd ),
                  1 : ( lambda: - params[0] * s2 * d2 ),
                  2 : ( lambda: math.pi * xdata * p3 ),
                  3 : ( lambda: p3 ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return ( "Etalon: f( x:p ) = p_0 / ( 1 + p_1 * sin^2( PI * x * p_2 + p_3 ) )" )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.

        Parameters
        ----------
        k : int
            the kth parameter.

        """
        if k == 0:
            return self.yUnit
        if k == 1:
            return units.Unit( "" )
        if k == 2:
            return units.Unit( units.si.rad ) / self.xUnit
        if k == 3:
            return units.Unit( units.si.rad )
        return self.yUnit





import numpy
import math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester


class CauchyPrior( Prior ):
    """
    Cauchy prior distribution.

        Pr( x ) =  scale / ( &pi; * ( scale^2 + ( x - center )^2 )

    By default: center = 0 and scale = 1.

    It can also have a limited domain. (To be done)
    By default the domain is [-Inf,+Inf].
    In computational practice it is limited to [-1e16, 1e16]

    domain2unit: u = arctan( ( d - c ) / s ) / pi + 0.5
    unit2domain: d = tan( ( u - 0.5 ) * pi ) * s + c

    Examples
    --------
    >>> pr = CauchyPrior()                         # center=0, scale=1
    >>> pr = CauchyPrior( center=1.0, scale=0.5 )
    >>> pr = CauchyPrior( limits=[0,None] )        # lowlimit=0, highlimit=inf
    >>> pr = CauchyPrior( center=1, circular=3 )   # circular between 0.5 and 2.5

    Attributes
    ----------
    center : float
        center of the Cauchy prior
    scale : float
        scale of the Cauchy prior

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain

    lowLimit and highLimit cannot be used in this implementation.

    """

    #  *********CONSTRUCTOR***************************************************
    def __init__( self, center=0.0, scale=1, limits=None, circular=False, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        center : float
            of the prior
        scale : float
            of the prior
        limits : None or [float,float]
            None    no limits are set
            2 floats    lowlimit and highlimit
        circular : bool or float
            bool : y|n circular with period from limits[0] to limits[1]
            float : period of circularity
        prior : CauchyPrior
            prior to copy (with new scale if applicable)

        """
        self.center = center
        self.scale = scale

        super( CauchyPrior, self ).__init__( limits=limits, circular=circular, prior=prior )

    def copy( self ):
        return CauchyPrior( prior=self, center=self.center, scale=self.scale,
                            limits=self.limits, circular=self.circular )

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Cauchy distribution.

        domain2unit: u = arctan( ( d - c ) / s ) / pi + 0.5

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return numpy.arctan( ( dval - self.center ) / self.scale ) / math.pi + 0.5

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Cauchy distribution.

        unit2domain: d = tan( ( u - 0.5 ) * pi ) * s + c

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        return numpy.tan( ( uval - 0.5 ) * math.pi ) * self.scale + self.center

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        xc = x - self.center
        return self.scale / ( ( self.scale * self.scale + xc * xc ) * math.pi )

# logResult has no better definition than the default: just take the math.log of result.
# No specialized method here.

    def partialLog( self, x ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        x : float
            the value

        """
        xc = x - self.center
        return - 2 * xc / ( self.scale * self.scale + xc * xc )

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def shortName( self ):
        """ Return a string representation of the prior.  """
        return str( "CauchyPrior" )






import numpy as numpy
from astropy import units
from astropy.modeling import Model as AstroModel
import math
import warnings
from . import Tools
from .Tools import setAttribute as setatt

from .Model import Model

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2021        Do Kester


class UserModel( Model ):
    """
    Wrapper for a Models where the User provides the method the evaluate
    f( x: p ) as userResult( x, param )

    and optionally the partial derivatives to p and x and a name

    df/dp as  userPartial( x, param )
    df/dx as  userDeriv( x. param )
    name  as  string


    Examples
    --------
    import numpy as np
    >>> def ur( x, p ) :
    >>>     return p[0] * np.sin( p[1] * x + p[2] * np.log( x + p[3] ) ) 
    >>> 
    >>>
    >>> mdl = UserModel( 4, ur, userName="slowchirp" )
    >>> print( mdl )
    UserModel( slowchirp )
    >>> print( mdl.npars )
    4


    Attributes
    ----------
    userResult : callable
        method to evaluate the result of the F(x:p)
        defined as method( x, p )
    userPartial : callable
        method to evaluate the partial derivatives df/dp
        defined as method( x, p ). Default: numeric derivative
    userDerivative : callable
        method to evaluate the derivative df/dx
        defined as method( x, p ). Default: numeric derivative
    userName : str
        name of the class

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """

    def __init__( self, npars, userResult, ndim=1, userPartial=None, userDeriv=None, 
                  userName="unknown", copy=None, **kwargs ):
        """
        User provided model.

        Parameters
        ----------
        npars : int
            number of parameters of this model
        userResult : callable
            method of the form userResult( x, p )
            where x is the independent variable; array_like
                  p is the parameter vector; array_like 
        ndim : int
            number of input streams.
        userPartial : callable
            method of the form userPartial( x, p )
            where x is the independent variable; array_like
                  p is the parameter vector; array_like 
        userDeriv : callable
            method of the form userDeriv( x, p )
            where x is the independent variable; array_like
                  p is the parameter vector; array_like 
        userName : str
            Name for this model
        copy : UserModel
            to be copied

        """

        self.setMethod( "userResult", userResult )

        super( ).__init__( npars, ndim=ndim, copy=copy, **kwargs )

        if copy is not None :
            setatt( self, "userPartial", copy.userPartial )
            setatt( self, "userDerivative", copy.userDerivative )
            setatt( self, "userName", copy.userName, type=str )
        else :
            self.setMethod( "userPartial", userPartial, numeric=self.numPartial )
            self.setMethod( "userDerivative", userDeriv, numeric=self.numDerivative )
            setatt( self, "userName", userName, type=str )

    def copy( self ):
        """ Copy method.  """
        return UserModel( self.npars, self.baseResult, ndim=self.ndim, copy=self )

    def setMethod( self, name, userMethod, numeric=None ) :
        if callable( userMethod ) :
            setatt( self, name, userMethod )
        elif callable( numeric ) :
            setatt( self, name, numeric )
            warnings.warn( "Using numeric approximations for %s" % name )
        else :
            raise ValueError( "%s %s is not callable" % ( name, str( userMethod ) ) )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return self.userResult( xdata, params )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        if parlist is None :
            return self.userPartial( xdata, params )
        else :
            return self.userPartial( xdata, params )[parlist]

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return self.userDerivative( xdata, params )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "UserModel( %s )" % self.userName )






import numpy as numpy
from . import Tools
from .NonLinearModel import NonLinearModel
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2021 Do Kester

class ExpModel( NonLinearModel ):
    """
    Exponential Model.

        f( x:p ) = p_0 * exp( p_1 * x )

    where p_0 = amplitude
          p_1 = slope
    As always x = input.

    The parameters are initialized at {1.0, -1.0}. It is a non-linear model.

    Beware of a positive 2nd parameter; when positive the model is going off
    to Infinity very quickly.

    When decay is True the model changes into a decay model:

        f( x:p ) = p_0 * exp( - p_1 * x )

    The parameters are initialized at {1.0, 1.0}.

    Attributes
    ----------
        sign : [-1,1]
        Whether decay is True or False.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Examples
    --------
    >>> em = ExpModel( )
    >>> print( em.getNumberOfParameters( ) )
    2

    Author  Do Kester

    """
    def __init__( self, decay=False, copy=None, **kwargs ):
        """
        Exponential model.
        <br>
        Number of parameters is 2.

        Parameters
        ----------
        decay : boolean
            changes sign of parameter[1]
        copy : ExpModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        if decay :
            setatt( self, "sign", -1 )
            param = [1.0,1.0]
            names = ["amplitude","decay"]
        else :
            setatt( self, "sign", 1 )
            param = [1.0,-1.0]
            names = ["amplitude","slope"]

        super( ExpModel, self ).__init__( 2, copy=copy, params=param,
                names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return ExpModel( decay=(self.sign==-1), copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return params[0] * numpy.exp( self.sign * params[1] * xdata )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata ), np ) )

        e = numpy.exp( self.sign * params[1] * xdata )

        parts = { 0 : ( lambda: e ),
                  1 : ( lambda: self.sign * params[0] * e * xdata ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative df/dx at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return self.sign * params[1] * self.baseResult( xdata, params )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "Exp: f( x:p ) = p_0 * exp( %sp_1 * x )" % ( "" if self.sign == 1 else "- " ) )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters: k    parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit ** ( -1 )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt

from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.8.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2021 Do Kester

class SineAmpModel( LinearModel ):
    """
    Sine with fixed frequency.

    Find amplitudes/phases for sinusoidal of a given frequency.

        f( x:p ) = p_0 cos( 2 &pi; &omega; x ) + p_1 sin( 2 &pi; &omega; x )

    where
        p_0 = amplitude of cosine
        p_1 = amplitude of sine
        &omega; is the fixed frequency of the model.

    It is a linear model with 2 parameters.

    Examples
    --------
    >>> sine = SineAmpModel( 150 )        # fixed frequency of 150 Hz

    Attributes
    ----------
    frequency : float
        the (fixed) frequency of the sinusoidal.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Alternate
    ---------
    This model is equivalent to
        SineModel( fixed={0:frequency} )

    """
    def __init__( self, frequency, copy=None, **kwargs ):
        """
        Sine model of a fixed frequency.

        Number of parameters is 2.

        Parameters
        ----------
        frequency : float
            the frequency
        copy : SineAmpModel
            model to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        names = ["cosamp", "sinamp"]
        super( SineAmpModel, self ).__init__( 2, copy=copy, names=names, **kwargs )

        if copy is None :
            self.frequency = frequency
        else :
            self.frequency = copy.frequency

    def copy( self ):
        """ Copy method.  """
        return SineAmpModel( self.frequency, copy=self )

    def __setattr__( self, name, value ) :
        if name == 'frequency' :
            setatt( self, name, value, type=float )
        else :
            super( SineAmpModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model (ignored in LinearModels)
        parlist : array_like
            list of indices of active parameters

        """
        nxdata = Tools.length( xdata )
        np = self.npmax if parlist is None else len( parlist )
        partial = numpy.zeros( ( nxdata, np ), dtype=float )

        parts = { 0 : ( lambda: numpy.cos( 2 * math.pi * self.frequency * xdata ) ),
                  1 : ( lambda: numpy.sin( 2 * math.pi * self.frequency * xdata ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        tpf = 2 * math.pi * self.frequency
        return tpf * ( params[1] * numpy.cos( tpf * xdata ) -
                       params[0] * numpy.sin( tpf * xdata ) )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return ( "SineAmp: f( x:p ) = p_0 * cos( 2 pi * x * f ) + " +
                 "p_1 * sin( 2 pi * x * f ); f = %f"%self.frequency )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        return self.yUnit





import numpy as numpy

from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2020 Do Kester


class Modifiable( object ):
    """
    Class adjoint to Model which implements the modifiable behaviour of some Models.

    In the inhertance list is should be *before* Model as it changes the behaviour of Model.

    """
    def __init__( self, modifiable=True ) :
        """
        Constructor for Modifiable

        Parameters
        ----------
        modifiable: bool
            True if the Model is to be considered modifiable.
        """

        setatt( self, "modifiable", modifiable )


    def isModifiable( self ) :
        return self.modifiable


    def vary( self, location=None, rng=None, **kwargs ) :
        """
        Vary the structure of a Modifiable Model
        Default implementation: does nothing.

        Parameters
        ----------
        location : int
            index of the item to be modified; otherwise random
        rng : RNG
            random number generator
        kwargs : keyword arguments
            for specific implementations
        """
        return True







#run with: python3 -m unittest TestMonteCarlo

import unittest
import os
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE

from astropy import units
import math
import matplotlib.pyplot as pyplot


from BayesicFitting import *
from BayesicFitting import formatter as fmt


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2016 Do Kester

class TestMonteCarlo( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testMonteCarlo1( self ):
        print( "====== MonteCarlo 1 ===================" )

        N = 100
        ran = numpy.random
        ran.seed( 12345 )
        noise = ran.standard_normal( N )
        x = numpy.arange( N, dtype=float ) - 3
        nn = 0.1
        for k in range( 5 ) :
            y = noise * nn

            m = PolynomialModel( 0 )
            ftr = Fitter( x, m )
            par = ftr.fit( y )
            std = ftr.getStandardDeviations()
            chisq = ftr.chisq

            mc = MonteCarlo( x, m, ftr.covariance )
            mc.mcycles = 1000
            lmce = ftr.monteCarloError( monteCarlo=mc )

            print( "noise  : ", fmt( nn ), "===========================================" )
            print( "params : ", fmt( par, format="%8.5f" ) )
            print( "stdevs : ", fmt( std, format="%8.5f" ) )
            print( "scale  : ", fmt( ftr.scale, format="%8.5f" ), fmt( nn ) )
            print( "chisq  : ", fmt( chisq, format="%8.5f" ),
                fmt( mc._eigenvalues, format="%8.5f" ), fmt( mc._eigenvectors, format="%8.5f" ) )
            print( "covar  : ", fmt( ftr.covariance, format="%8.5f" ) )
            print( "mcerr  : ", fmt( lmce[0], format="%8.5f" ) )
            self.assertTrue( abs( std[0] - lmce[0] ) < 0.1 * std[0] )
            self.assertTrue( par[0] < 0.05 * nn )
            nn *= 10



#        self.assertAlmostEqual( par, 0.0 )
#        self.assertAlmostEqual( par, 0.1 )
#        self.assertAlmostEqual( chisq, 0.857142857143 )

    def testMonteCarlo2( self ):
        print( "====== MonteCarlo 2 ===================" )

        x = numpy.arange( 7, dtype=float ) - 3
        y = numpy.asarray( [-1,-1,-1,0,1,1,1], dtype=float )
        m = PolynomialModel( 1 )
        ftr = Fitter( x, m )
        par = ftr.fit( y )
        std = ftr.getStandardDeviations()
        yfit = m.result( x )
        chisq = ftr.chisq
        hes = ftr.hessian
#        mc = MonteCarlo( x, m, chisq, hessian=hes )
        mc = MonteCarlo( x, m, ftr.covariance )
        mc.mcycles = 1000
        lmce = ftr.monteCarloError( monteCarlo=mc )

        print( "params : ", par )
        print( "stdevs : ", std )
        print( "scale  : ", ftr.getScale() )
        print( "chisq  : ", chisq )
        print( "evals  : ", mc._eigenvalues )
        print( mc._eigenvectors )
        print( "hessi  :\n", ftr.hessian )
        print( "covar  :\n", ftr.covariance )
        print( "mcerr  : ", lmce )

        numpy.testing.assert_array_almost_equal( par, numpy.asarray([0.0,0.42857142857142855]) )
#        numpy.testing.assert_array_almost_equal( std, numpy.asarray([0.1564921592871903,0.07824607964359515]) )
        self.assertAlmostEqual( chisq, 0.857142857143 )

        if self.doplot :
            pyplot.plot( x, y, 'k*' )
            pyplot.plot( x, yfit, 'g-' )
            pyplot.plot( x, yfit+lmce, 'r-' )
            pyplot.plot( x, yfit-lmce, 'r-' )
            pyplot.show()


    def testMonteCarlo3( self ):
        print( "====== MonteCarlo 3 ===================" )

        N = 101
        x = numpy.arange( N, dtype=float ) * 0.1
        ran = numpy.random
        ran.seed( 1235 )
        noise = ran.standard_normal( N )

        ym = x * x + 0.03 * x + 0.05
        y1 = ym + 10 * noise

        pm = PolynomialModel( 2 )

        ftr = Fitter( x, pm )

        pars1 = ftr.fit( y1 )
        stdv1 = ftr.getStandardDeviations( )
        print( "parameters : ", pars1 )
        print( "std devs   : ", stdv1 )
        print( "chisquared : ", ftr.chisq )

        lmce = ftr.monteCarloError( )
        chisq = ftr.chisq

        mce = MonteCarlo( x, pm, ftr.covariance )
        mce1 = mce.getError( )
        assertAAE( lmce, mce1 )

        yfit = pm.result( x )
        s2 = numpy.sum( numpy.square( ( yfit - ym ) / lmce ) )
        print( s2, math.sqrt( s2 / N ) )

        integral = numpy.sum( yfit )
        s1 = 0
        s2 = 0
        k = 0
        for k in range( 1, 100001 ):
            rv = mce.randomVariant( x )
            s1 += numpy.sum( rv )
            s2 += numpy.sum( numpy.square( rv - yfit ) )
            if k % 10000 == 0:
                print( "%6d  %10.3f %10.3f %10.3f" % ( k, integral, s1/k, math.sqrt( s2/k ) ) )

        ### TBC  dont know why the factor 1000 is there. ########
        print( abs( integral - s1/k ), math.sqrt( s2/( k * 1000 ) ) )
        self.assertTrue( abs( integral - s1/k ) < math.sqrt( s2/( k * 1000 ) ) )

        if self.doplot :
            pyplot.plot( x, y1, 'b.' )

            pyplot.plot( x, ym, 'k-' )
            pyplot.plot( x, yfit, 'g-' )
            pyplot.plot( x, yfit+lmce, 'r-' )
            pyplot.plot( x, yfit-lmce, 'r-' )
            pyplot.show()

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestMonteCarlo.__class__ )


if __name__ == '__main__':
    unittest.main()





from __future__ import print_function

import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt
from . import Plotter
import sys
import warnings
import matplotlib.pyplot as plt

from .Explorer import Explorer
from .Model import Model
from .Walker import Walker
from .WalkerList import WalkerList
from .Sample import Sample
from .SampleList import SampleList

from .Problem import Problem
from .ClassicProblem import ClassicProblem
from .ErrorsInXandYProblem import ErrorsInXandYProblem
from .MultipleOutputProblem import MultipleOutputProblem
from .EvidenceProblem import EvidenceProblem

from .ErrorDistribution import ErrorDistribution
from .ScaledErrorDistribution import ScaledErrorDistribution
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .UniformErrorDistribution import UniformErrorDistribution
from .ExponentialErrorDistribution import ExponentialErrorDistribution
from .ModelDistribution import ModelDistribution
from .BernoulliErrorDistribution import BernoulliErrorDistribution

from .Engine import Engine
from .StartEngine import StartEngine
from .ChordEngine import ChordEngine
from .GibbsEngine import GibbsEngine
from .GalileanEngine import GalileanEngine
from .StepEngine import StepEngine
## for Dynamic Models import the classes
from .BirthEngine import BirthEngine
from .DeathEngine import DeathEngine
## for Modifiable Models:
from .StructureEngine import StructureEngine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2022 Do Kester

class NestedSampler( object ):
    """
    Nested Sampling is a novel technique to do Bayesian calculations.

    Nested Sampling calculates the value of the evidence while it
    simultaneously produces samples from the posterior probability
    distribution of the parameters, given a Model and a
    dataset ( x-values, y-values and optionally weights ).
    The samples are collected in a SampleList.

    NestedSampler is constructed according to the ideas of John Skilling
    and David MacKay ( ref TBD ).

    Internally it contains an ensemble ( by default: 100 ) of trial samples,
    called walkers. Initially they are randomly distributed over the space
    available to the parameters. This randomness is distributed according
    to the prior distribution of the parameters. In most simple cases it
    will be uniform.

    In an iterative process, the sample with the lowest likelihood is
    selected and replaced by a copy of one of the others.
    The parameters of the copy are randomly walked around under the
    condition that its likelihood stays larger than the selected lowest likelihood.
    A new independent trial sample is constructed.
    The original lowest sample is placed, appropriately weighted, into the
    SampleList for output.

    This way the likelihood is climbed until the maximum is found.
    Along the way the integral of the likelihood function is
    calculated, which is equal to the evidence for the model,
    given the dataset.

    There are different likelihood functions available: Gaussian (Normal),
    Laplace (Norm-1), Uniform (Norm-Inf), Poisson (for counting processes),
    Bernoulli (for categories), Cauchy (aka Lorentz) and
    Exponential (generalized Gaussian).
    A mixture of ErrorDistributions can also be defined.
    By default the Gaussian distribution is used.

    Except for Poisson and Bernoulli , all others are ScaledErrorDistributions.
    The scale is a HyperParameter which can either be fixed, by setting the
    attribute errdis.scale or optimized, given the model parameters and the data.
    By default it is optimized for Gaussian and Laplace distributions.
    The prior for the noise scale is a JeffreysPrior.

    The Exponential also has a power as HyperParameter which can be fixed
    or optimized. Its default is 2 (==Gaussian).

    For the randomization of the parameters within the likelihood constraint
    so-called engines are written.
    By default only engines 1 and 2 is switched on.

    1. GalileanEngine.
        It walks all (super)parameters in a fixed direction for about 5 steps.
        When a step ends outside the high likelihood region the direction is
        mirrored on the lowLikelihood edge and continued.

    2. ChordEngine.
        It draws a randomly oriented line through a point inside the region,
        until it reaches outside the restricted likelihood region on both sides.
        A random point is selected on the line until it is inside the likelihood region.
        This process runs several times

    3. GibbsEngine.
        It moves each of the parameters by a random step, one at a time.
        It is a randomwalk.

    4. StepEngine.
        It moves all parameters in a random direction. It is a randomwalk.

    5. CrossEngine.
        The parameters of 2 walkers are mixed in a random way.

    For dynamic models 2 extra engines are defined:

    6. BirthEngine.
        It tries to increase the number of parameters.
        It can only be switched on for Dynamic Models.

    7. DeathEngine.
        It tries to decrease the number of parameters.
        It can only be switched on for Dynamic Models.

    For modifiable models 1 engine is defined.

    8. StructureEngine.
        It alters the internal structure of the model.
        It can only be switched on for Modifiable Models.

    Attributes
    ----------
    xdata : array_like
        array of independent input values
    model : Model
        the model function to be fitted
    ydata : array_like
        array of dependent (to be fitted) data
    weights : array_like (None)
        weights pertaining to ydata
    problem : Problem (ClassicProblem)
        to be solved (container of model, xdata, ydata and weights)
    distribution : ErrorDistribution
        to calculate the loglikelihood
    ensemble : int (100)
        number of walkers
    discard : int (1)
        number of walkers to be replaced each generation
    rng : RandomState
        random number generator
    seed : int (80409)
        seed of rng
    rate : float (1.0)
        speed of exploration
    maxsize : None or int
        maximum size of the resulting sample list (None : no limit)
    minimumIterations : int (100)
        minimum number of iterations (adapt when starting problems occur)
    end : float (2.0)
        stopping criterion
    verbose : int
        level of blabbering
    walkers : WalkerList
        ensemble of walkers that explore the likelihood space
    samples : SampleList
        Samples resulting from the exploration
    engines : list of Engine
        Engine that move the walkers around within the given constraint: logL > lowLogL
    initialEngine : Engine
        Engine that distributes the walkers over the available space
    restart : StopStart (TBW)
        write intermediate results to (optionally) start from.

    Author       Do Kester.

    """
    TWOP31 = 2 ** 31        ## 2 ** 32 raises an error on Windows systems.
    ENSEMBLE = 100

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata=None, model=None, ydata=None, weights=None,
                problem=None, distribution=None, limits=None, keep=None, ensemble=100,
                discard=1, seed=80409, rate=1.0, engines=None, maxsize=None,
                threads=False, verbose=1 ) :
        """
        Create a new class, providing inputs and model.

        Either (model,xdata,ydata) needs to be provided or a completely filled
        problem.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
            the model needs priors for the parameters and (maybe) limits
        ydata : array_like
            array of dependent (to be fitted) data
        weights : array_like (None)
            weights pertaining to ydata
        problem : None or string or Problem
            Defines the kind of problem to be solved.

            None        same as "classic"
            "classic" 	ClassicProblem
            "errors"	ErrorsInXandYProblem
            "multiple"	MultipleOutputProblem

            Problem     Externally defined Problem. When Problem has been provided,
                        xdata, model, weights and ydata are not used.
        keep : None or dict of {int:float}
            None of the model parameters are kept fixed.
            Dictionary of indices (int) to be kept at a fixed value (float).
            Hyperparameters follow model parameters.
            The values will override those at initialization.
            They are used in this instantiation, unless overwritten at the call to sample()
        distribution : None or String or ErrorDistribution
            Defines the ErrorDistribution to be used
            When the hyperpar(s) are not to be kept fixed, they need `Prior` and maybe limits.

            None            same as "gauss"
            "gauss"         GaussErrorDistribution with (fixed) scale equal to 1.0
            "laplace"       LaplaceErrorDistribution with 1 hyperpar scale
            "poisson"       PoissonErrorDistribution no hyperpar
            "cauchy"        CauchyErrorDstribution with 1 hyperpar scale
            "uniform"       UniformErrorDistribution with 1 hyperpar scale
            "bernoulli"     BernoulliErrorDistribution no hyperpar
            "exponential"   ExponentialErrorDistribution with 2 hyperpar (scale, power)

            ErrorDistribution Externally defined ErrorDistribution
        limits : None or [low,high] or [[low],[high]]
            None    no limits implying fixed hyperparameters of the distribution
            low     low limit on hyperpars
            high    high limit on hyperpars
            When limits are set the hyperpars are not fixed.
        ensemble : int
            number of walkers
        discard : int
            number of walkers to be replaced each generation
        seed : int
            seed of random number generator
        rate : float
            speed of exploration
        engines : None or (list of) string or (list of) Engine
            to randomly move the walkers around, within the likelihood bound.

            None        use a Problem defined selection of engines
            "galilean"  GalileanEngine	move forward and mirror on edges
            "chord"     ChordEngine   	select random point on random line
            "gibbs" 	GibbsEngine 	move one parameter at a time
            "step"  	StepEngine    	move all parameters in arbitrary direction

            For Dynamic models only:
            "birth" 	BirthEngine     increase the parameter list of a walker by one
            "death" 	DeathEngine     decrease the parameter list of a walker by one

            For Modifiable models only:
            "struct"    StructureEngine change the (internal) structure.

            Engine      an externally defined (list of) Engine
        maxsize : None or int
            maximum size of the resulting sample list (None : no limit)
        threads : bool (False)
            Use Threads to distribute the diffusion of discarded samples over the available cores.
        verbose : int (1)
            0   silent
            1   basic information
            2   more about every 100th iteration
            3   more about every iteration
            >4  for debugging

        """
        if problem is None :
            problem = "classic"

        self.setProblem( problem, model=model, xdata=xdata, ydata=ydata, weights=weights )

        if model is None :
            model = self.problem.model

        if model is not None and not model.hasPriors() and model.npars > 0:
            warnings.warn( "Model needs priors and/or limits" )

        self.keep = keep

        if ensemble is not None :
            self.ensemble = ensemble

        self.discard = discard

        self.rng = numpy.random.RandomState( seed )
        self.seed = seed
        self.maxsize = maxsize
        object.__setattr__( self, "verbose", verbose )
        self.rate = rate
        self.restart = None                     ## TBD

        self.minimumIterations = 100
        self.end = 2.0
        self.maxtrials = 5
        self.threads = threads

        self.iteration = 0

        if distribution is not None :
            self.setErrorDistribution( distribution, limits=limits )
        elif limits is not None :
            self.setErrorDistribution( self.problem.myDistribution(), limits=limits )
        else :
            self.setErrorDistribution( self.problem.myDistribution(), scale=1.0 )

        self.setEngines( engines )

        ## Initialize the sample list
        self.samples = SampleList( model, 0, ndata=self.problem.ndata )


    #  *******SAMPLE************************************************************
    def sample( self, keep=None, plot=False ):
        """
        Sample the posterior and return the 10log( evidence )

        A additional result of this method is a SampleList which contains
        samples taken from the posterior distribution.

        Parameters
        ----------
        keep : None or dict of {int:float}
            Dictionary of indices (int) to be kept at a fixed value (float)
            Hyperparameters follow model parameters
            The values will override those at initialization.
            They are only used in this call of fit.
        plot : bool or str
            bool    show a plot of the final results
            "iter" 	show iterations
            "all"  	show iterations and final result
            "last" 	show final result

        """
        keep = self.initSample( keep=keep )

        tail = self.initReport( keep=keep, plot=plot )

        explorer = Explorer( self, threads=self.threads )

        self.logZ = -sys.float_info.max
        self.info = 0
        self.logWidth = math.log( 1.0 - math.exp( -1.0 / self.ensemble ) )


#        TBD put self.logWidth in the saved file too
#        if self.optionalRestart() :
#            self.logWidth -= self.iteration * ( 1.0 * self.discard ) / self.ensemble

        ## iterate until done
        while self.iteration < self.getMaxIter( ):

            self.walkers.sort( key=self.walkerLogL )    # sort the walker list on logL

            self.calculateUnitRange()

            worst = self.worst                      # the worst are low in the sorted ensemble
            self.lowLhood = self.walkers[worst].logL

            self.updateEvidence( worst )            # Update Z and H and store posterior samples

            self.iterReport( worst - 1, tail, plot=plot ) # some output when needed

            self.samples.weed( self.maxsize )       # remove overflow in samplelist

            self.iteration += 1

            self.updateWalkers( explorer, worst )

            self.optionalSave( )


        # End of Sampling: Update and store the remaining walkers
        self.updateEvidence( self.ensemble )        # Update Evidence Z and Information H

        # Calculate weighted average and stdevs for the parameters;
        self.samples.LogZ = self.logZ
        self.samples.info = self.info
        self.samples.normalize( )

        # put the info into the model
        if self.problem.model and not self.problem.model.isDynamic() :
            self.problem.model.parameters = self.samples.parameters
            self.problem.model.stdevs = self.samples.stdevs

        self.lastReport( -1, plot=plot )

        return self.evidence

    def initSample( self, ensemble=None, keep=None ) :

        if keep is None :
            keep = self.keep
        fitIndex, allpars = self.makeFitlist( keep=keep )

        if ensemble is None :
            ensemble = self.ensemble
        self.initWalkers( ensemble, allpars, fitIndex )

        for eng in self.engines :
            eng.walkers = self.walkers
            eng.lastWalkerId = len( self.walkers )

        self.distribution.ncalls = 0                      #  reset number of calls

        return keep

    def sortLogL( self, walkers ) :
        return numpy.argsort( [w.logL for w in walkers] )

    def walkerLogL( self, w ) :
        return w.logL

    def makeFitlist( self, keep=None ) :
        """
        Make list of indices of (hyper)parameters that need fitting.

        Parameters
        ----------
        keep : None or dict of {int : float}
            dictionary of indices that need to be kept at the float value.
        """
        allpars = numpy.zeros( self.problem.npars + self.distribution.nphypar,
                               dtype=self.problem.partype )

        np = self.problem.npars
        fitlist = [k for k in range( np )]
        nh = -self.distribution.nphypar
        for sp in self.distribution.hyperpar :
            if not sp.isFixed and sp.isBound() :
                fitlist += [nh]                             # it is to be optimised
            else :
                allpars[nh] = self.distribution.hypar[nh]   # fill in the fixed value

            nh += 1

        nh = self.distribution.nphypar
        if keep is not None :
            fitl = []
            kkeys = list( keep.keys() )
            for k in range( np + nh ) :
                if k in kkeys :
                    allpars[k] = keep[k]                    # save keep.value in allpars
                elif fitlist[k] in kkeys :
                    allpars[k] = keep[fitlist[k]]           # save keep.value in allpars
                else :
                    fitl += [fitlist[k]]                    # list of pars to be fitted
            fitlist = fitl

        return ( fitlist, allpars )

    def calculateUnitRange( self ) :
        self.engines[0].calculateUnitRange()
        for eng in self.engines :
            eng.unitRange = self.engines[0].unitRange
            eng.unitMin   = self.engines[0].unitMin


    def doIterPlot( self, plot ) :
        if isinstance( plot, str ) :
            return plot == 'iter' or plot == 'all'
        else :
            return False

    def doLastPlot( self, plot ) :
        if isinstance( plot, str ) :
            return plot == 'last' or plot == 'all'
        else :
            return plot == True


    def initReport( self, keep=None, plot=False ) :

        self.plotData( plot=self.doIterPlot( plot ) )

        if self.verbose >= 1 :
            fitIndex = self.walkers[0].fitIndex
            print( "Fit", ( "all" if keep is None
                                  else fitIndex ), "parameters of" )
            if self.problem.model :
                print( " ", self.problem.model._toString( "  " ) )
            else :
                print( " ", self.problem )
            print( "Using a", self.distribution, end="" )
            np = -1
            cstr = " with "
            for name,hyp in zip( self.distribution.PARNAMES, self.distribution.hyperpar ) :
                print( cstr, end="" )
                if np in fitIndex :
                    print( "unknown %s" % name, end="" )
                else :
                    print( "%s = %7.2f " % (name, hyp.hypar), end="" )
                np -= 1
                cstr = " and "
            print( "\nMoving the walkers with ", end="" )
            for eng in self.engines :
                print( " ", eng, end="" )
            print( "" )
            if self.threads :
                print( "Using threads." )

        tail = 0
        if self.verbose > 1 :
            while fitIndex is not None and fitIndex[-tail-1] < 0 : 
                tail += 1

#            tail = self.distribution.nphypar
            if tail == 0 :
                print( "Iteration     logZ        H       LowL     npar parameters" )
            else :
                print( "Iteration     logZ        H       LowL     npar parameters", end="" )
                for k in range( 1,  min( 5, len( fitIndex ) - tail ) ) :
                    print( "        ", end="" )
                for k in range( tail ) :
                    print( "   %s" %self.distribution.hyparname( k ), end="" )
                print( "" )

        return tail

    def iterReport( self, kw, tail, plot=False ) :

        if self.verbose >= 3 or ( self.verbose >= 1 and
                                  self.iteration % 100 == 0 ):
            if self.verbose == 1 :
                if ( self.iteration / 100 ) % 50 == 49 :
                    nwln = "\n"
                else :
                    nwln = ""
                print( ">", end=nwln, flush=True )
            else :
                pl = self.walkers[kw].allpars
                fi = self.walkers[kw].fitIndex
                if fi is not None : 
                    pl = pl[fi]
                np = len( pl )
#                tail = self.distribution.nphypar
                print( "%8d %#10.3g %8.1f %#10.3g %6d "%( self.iteration, self.logZ,
                    self.info, self.lowLhood, np ), fmt( pl, max=4, tail=tail ) )
#                print( "%-16.16s " % self.engines[1], end="" )
#                self.engines[1].printReport()


            self.plotResult( self.walkers[kw], self.iteration, plot=self.doIterPlot( plot ) )

    def lastReport( self, kw, plot=False ) :

        if self.verbose > 0 :
            if self.verbose == 1 :
                print( "\nIteration   logZ        H     LowL     npar    parameters" )
            pl = self.walkers[kw].allpars
            fi = self.walkers[kw].fitIndex
            if fi is not None : 
                pl = pl[fi]
            np = len( pl )
            print( "%8d %#10.3g %8.1f %#10.3g %6d "%( self.iteration, self.logZ,
                    self.info, self.lowLhood, np ) )
            print( fmt( pl, max=None ) )

        if self.verbose >= 1 :
            self.report()

        if self.doLastPlot( plot ) :
#            Plotter.plotFit( self.problem.xdata, self.problem.ydata, yfit=self.yfit, residuals=True )
            Plotter.plotSampleList( self.samples, self.problem.xdata, self.problem.ydata, 
                    figsize=[12,8], residuals=True )

    def getMaxIter( self ) :
        """
        Return the maximum number of iteration.
        """
        return max( self.minimumIterations, self.end * self.ensemble * self.info / self.worst )

    def getMaxIterMED( self ) :
        """
        Return the maximum number of iteration in case of EvidenceProblem.
        """
        try :
            self.csEvid += [math.exp( self.logZ )]
        except :
            self.csEvid = [math.exp( self.logZ )]

        if self.iteration < self.minimumIterations :
            return self.minimumIterations

        kmed = 0
        medlim = 0.5 * self.csEvid[-1]
        for cs in self.csEvid :
            if cs > medlim : break
            kmed += 1

        return self.end * kmed


    def getMaxIterEP( self ) :
        """
        Return the maximum number of iteration in case of EvidenceProblem.
        """
        if self.iteration < self.minimumIterations :
            return self.minimumIterations
        if ( self.samples[-1].logW < self.samples[-100].logW and 
             self.samples[-1].logW < self.logZ - 16 ) :
            return self.iteration
        else :
            return self.iteration + 100  

#   det getScale( self, walker ) :
#       np = walker.model.npchain
#       return self.distribution.getScale( walker.model, params=walker.allpars[:np] )

#  ===================================================================================
    def optionalRestart( self ):
        """
        Restart the session from a file. (not yet operational)
        """
        if self.restart is not None and self.restart.wantRestore( ):
            self.walkers, self.samples = self.restart.restore( self.walkers, self.samples )
            self.logZ = self.walkers.logZ
            self.info = self.walkers.info
            self.iteration = self.walkers.iteration
            return True
        return False

    def optionalSave( self ):
        """
        Save the session to a file. (not yet operational)
        """
        if self.restart is not None and self.restart.wantSave( ) and self.iteration % 100 == 0:
            self.walkers.logZ = self.logZ
            self.walkers.info = self.info
            self.walkers.iteration = self.iteration
            self.restart.save( self.walkers, self.samples )


    def updateEvidence( self, worst ) :
        """
        Updates the evidence (logZ) and the information (H)

        The walkers need to be sorted to logL

        Parameters
        ----------
        worst : int
            Number of walkers used in the update

        """

        for kw in range( worst ) :
            logWeight = self.logWidth + self.walkers[kw].logL

            # update evidence, logZ
            logZnew = numpy.logaddexp( self.logZ, logWeight )

            # update Information, H
            self.info = ( math.exp( logWeight - logZnew ) * self.lowLhood +
                    math.exp( self.logZ - logZnew ) * ( self.info + self.logZ ) - logZnew )

#            print( kw, fmt( self.walkers[kw].logL ), fmt( self.logWidth ), fmt( logWeight ), fmt( logZnew ), fmt( self.info ) )

            if math.isnan( self.info ) :
                self.info = 0.0
            self.logZ = logZnew

            # store posterior samples
            smpl = self.walkers[kw].toSample( logWeight )
            self.samples.add( smpl )

#            self.logWidth -= 1.0 / self.ensemble
            self.logWidth -= 1.0 / ( self.ensemble - kw )

#        self.logWidth -= worst / self.ensemble

        return

    def copyWalker( self, worst ):
        """
        Kill worst walker( s ) in favour of one of the others

        Parameters
        ----------
        worst : int
            number of Walkers to copy
        """
        for k in range( worst ) :
            kcp = self.rng.randint( worst, self.ensemble )
            self.walkers.copy( kcp, k )
            setatt( self.walkers[k], "parent", kcp )
            setatt( self.walkers[k], "start", self.iteration )


    def updateWalkers( self, explorer, worst ) :
        """
        Update the walkerlist in place.

        Parameters
        ----------
        explorer : Explorer
            Explorer object
        worst : int
            number of walkers to update
        """
        self.copyWalker( worst )

        # Explore the copied walker(s)
        wlist = [k for k in range( worst )]
        explorer.explore( wlist, self.lowLhood, self.iteration )


    #  *********INTERNALS***************************************************
    def __setattr__( self, name, value ) :

        # for hypar fitting
        if name == "scale" and isinstance( self.distribution, ScaledErrorDistribution ) :
            self.distribution.scale = value
        elif name == "power" and isinstance( self.distribution, ExponentialErrorDistribution ) :
            self.distribution.power = value
        else :
            object.__setattr__( self, name, value )
            if name == "verbose" :
                for eng in self.engines :
                    eng.verbose = value

    def __getattr__( self, name ) :

        if name == "ensemble" :
            return len( self.walkers )
        elif name == "xdata" :
            return self.problem.xdata
        elif name == "model" :
            return self.problem.model
        elif name == "ydata" :
            return self.problem.ydata
        elif name == "weights" :
            return self.problem.weights
        elif name == "worst" :
            return self.discard
        elif name == "evidence" :
            return self.logZ / math.log( 10.0 )
        elif name == "logZprecision" :
            return math.sqrt( self.info / self.ensemble )
        elif name == "precision" :
            return self.logZprecision / math.log( 10.0 )
        elif name == "information" :
            return self.info
        elif name == "parameters" :
            return self.samples.getParameters()
        elif name == "stdevs" or name == "standardDeviations" :
            self.samples.getParameters()
            return self.samples.stdevs
        elif name == "hypars" :
            return self.samples.hypars
        elif name == "stdevHypars" :
            return self.samples.stdevHypars
        elif name == "scale" :
            return self.samples.hypars[0]
        elif name == "stdevScale" :
            return self.samples.stdevHypars[0]
        elif name == "modelfit" or name == "yfit" :
            return self.samples.average( self.problem.xdata )

    #  *********DISTRIBUTIONS***************************************************

    def setProblem( self, name, model=None, xdata=None, ydata=None, weights=None ) :
        """
        Set the problem for this run.

        If name is a Problem, then the keyword arguments (xdata,model,ydata,weights)
        are overwritten provided they are not None

        Parameters
        ----------
        name : string or Problem
            name of problem
            Problem Use this one
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with ydata

        """
        problemdict = {
            "classic"  : ClassicProblem,
            "errors"   : ErrorsInXandYProblem,
            "multiple" : MultipleOutputProblem,
            "evidence" : EvidenceProblem
        }
#           "salesman" : SalesmanProblem
#           "order" : OrderProblem

        if isinstance( name, Problem ) :
            self.problem = name
            if model is not None : self.problem.model = model
            if xdata is not None : self.problem.xdata = numpy.asarray( xdata )
            if ydata is not None : self.problem.ydata = numpy.asarray( ydata )
            if weights is not None : self.problem.weights = numpy.asarray( weights )
            return

        if not isinstance( name, str ) :
            raise ValueError( "Cannot interpret ", name, " as string or Problem" )

        name = str.lower( name )
        try :
            myProblem = problemdict[name]
            self.problem = myProblem( model, xdata=xdata, ydata=ydata, weights=weights )
        except :
            raise ValueError( "Unknown problem name %s" % name )



    #  *********DISTRIBUTIONS***************************************************
    def setErrorDistribution( self, name=None, limits=None, scale=1.0, power=2.0 ):
        """
        Set the error distribution for calculating the likelihood.

        Parameters
        ----------
        name : None or string or ErrorDistribution
            None    distribution is Problem dependent.
            string  name of distribution; one of
                    "gauss", "laplace", "poisson", "cauchy", "uniform",
                    "exponential", "bernoulli"
            ErrorDistribution use this one
        limits : None or [low,high] or [[low],[high]]
            None    no limits implying fixed hyperparameters (scale,power,etc)
            low     low limit on hyperpars (needs to be >0)
            high    high limit on hyperpars
            when limits are set, the hyperpars are *not* fixed.
        scale : float
            fixed scale of distribution
        power : float
            fixed power of distribution

        """
        if name is None :
            name = self.problem.myDistribution()
        elif isinstance( name, ErrorDistribution ) :
            self.distribution = name
            return

        if not isinstance( name, str ) :
            raise ValueError( "Cannot interpret ", name, " as string or ErrorDistribution" )

        name = str.lower( name )
#        print( name )
        if name == "gauss" :
            self.distribution = GaussErrorDistribution( scale=scale, limits=limits )
        elif name == "laplace" :
            self.distribution = LaplaceErrorDistribution( scale=scale, limits=limits )
        elif name == "poisson" :
            self.distribution = PoissonErrorDistribution()
        elif name == "cauchy" :
            self.distribution = CauchyErrorDistribution( scale=scale, limits=limits )
        elif name == "uniform" :
            self.distribution = UniformErrorDistribution( scale=scale, limits=limits )
        elif name == "exponential" :
            self.distribution = ExponentialErrorDistribution( scale=scale, power=power, limits=limits )
        elif name == "model" :
            self.distribution = ModelDistribution( scale=scale, limits=limits )
#            self.getMaxIter = self.getMaxIterEP
#            self.minimumIterations = max( self.minimumIterations, 100 )
        elif name == "bernoulli" :
            self.distribution = BernoulliErrorDistribution()
        else :
            raise ValueError( "Unknown error distribution %s" % name )

    def setEngines( self, engines=None, enginedict=None ) :
        """
        initialize the engines.

        Parameters
        ----------
        engines : None or [Engine] or [string]
            None       engines is Problem dependent
            [Engines]  list of Engines to use
            [string]   list engine names
        enginedict : dictionary of { str : Engine }
            connecting names to the Engines
        """

        if enginedict is None :
            enginedict = {
                "galilean" : GalileanEngine,
                "chord" :    ChordEngine,
                "birth" : 	 BirthEngine,
                "death" : 	 DeathEngine,
                "struct": 	 StructureEngine,
                "gibbs" : 	 GibbsEngine,
                "step"  : 	 StepEngine }

        if engines is None :
            engines = self.problem.myEngines()

        self.engines = []
        usePhantoms = self.step is not None

        if isinstance( engines, str ) :
            engines = [engines]
        for name in engines :
            if isinstance( name, Engine ) :
                engine = name
                engine.members = self.walkers
                engine.errdis = self.distribution
                engine.verbose = self.verbose
                if usePhantoms :
                    engine.setWalker = engine.setWalkerAdd2List
                self.engines += [engine]
                continue

            if not isinstance( name, str ) :
                raise ValueError( "Cannot interpret ", name, " as string or as Engine" )

            try :
                Eng = enginedict[name]
                seed = self.rng.randint( self.TWOP31 )
                engine = Eng( self.walkers, self.distribution,
                              seed=seed, verbose=self.verbose )
            except :
                raise ValueError( "Unknown Engine name : %10s" % name )

            self.engines += [engine]


    #  *********INITIALIZATION***************************************************
    def initWalkers( self, ensemble, allpars, fitIndex, startdict=None ):
        """
        Initialize the walkers at random values of parameters and scale

        Parameters
        ----------
        ensemble : int
            length of the walkers list
        allpars : array_like
            array of (hyper)parameters
        fitIndex : array_like
            indices of allpars to be fitted
        startdict : dictionary of { str : Engine }
            connecting a name to a StartEngine
        """
        if startdict is None :
            startdict = { "start" : StartEngine }

        # Make the walkers list
        self.walkers = WalkerList( self.problem, ensemble, allpars, fitIndex )

        if self.initialEngine is not None:
            # decorate with proper information
            self.initialEngine.members = self.walkers
            self.initialEngine.errdis = self.distribution

        else :
            try :
                name = self.problem.myStartEngine()
                StartEng = startdict[name]
                seed = self.rng.randint( self.TWOP31 )
                self.initialEngine = StartEng( self.walkers, self.distribution,
                        seed=seed )
            except :
                raise ValueError( "Unknown StartEngine name : %10s" % name )

        # Calculate logL for all walkers.
        self.distribution.lowLhood = -math.inf
        for walker in self.walkers :
            self.initialEngine.execute( walker.id, -math.inf )

#        for w in self.walkers :
#            print( w.id, w.problem.model.npars, len( w.problem.model.parameters), len( w.allpars) )
#            w.check( nhyp=self.distribution.nphypar )

    def plotData( self, plot=False ):
        """
        Initialize the plot with data.

        Parameters
        ----------
        plot : bool
            whether to plot.
        """
        if not plot :
            return
        plt.figure( 'iterplot' )
        plt.plot( self.problem.xdata, self.problem.ydata, 'k.' )
        plt.show( block=False )

    def plotResult( self, walker, iter, plot=False ):
        """
        Plot the results for a walker.

        Parameters
        ----------
        walker : Walker
            the walker to plot
        iter : int
            iteration number
        plot : bool
            whether to plot.
        """
        if not plot :
            return

        plt.figure( 'iterplot' )
        if self.line is not None :
            ax = plt.gca()
            plt.pause( 0.02 )
            ax.lines.remove( self.line )
            self.text.set_text( "Iteration %d" % iter )
        else :
            self.text = plt.title( "Iteration %d" % iter )
            self.ymin, self.ymax = plt.ylim()

        param = walker.allpars
        model = walker.problem.model
        mock = model.result( self.problem.xdata, param )
        self.line, = plt.plot( self.problem.xdata, mock, 'r-' )
        dmin = min( self.ymin, numpy.min( mock ) )
        dmax = max( self.ymax, numpy.max( mock ) )
        plt.ylim( dmin, dmax )
        plt.show( block=False )


    def report( self ):
        """
        Final report on the run.
        """
#        print( "Rate        %f" % self.rate )
        print( "Engines              success     reject     failed       best      calls" )

        for engine in self.engines :
            print( "%-16.16s " % engine, end="" )
            engine.printReport()
        print( "Calls to LogL     %10d" % self.distribution.ncalls, end="" )
        if self.distribution.nparts > 0 :
            print( "   to dLogL %10d" % self.distribution.nparts )
        else :
            print( "" )

        print( "Samples  %10d" % len( self.samples ) )
        print( "Evidence    %10.3f +- %10.3f" % (self.evidence, self.precision ) )






import math as math
import random as random

from .Tools import setAttribute as setatt
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class ExponentialPrior( Prior ):
    """
    Exponential prior distribution.

        Pr( x ) = exp( -x / scale )

    By default scale = 1.

    The domain is [0,+Inf].
    In computational practice the domain is limited to about [0,36] scale units

    domain2unit: u = 1 - exp( - d / scale )
    unit2domain: d = -log( 1 - u ) * scale

    Optionally one can set a zero fraction as the fraction of u where the unit2Domain
    method returns a 0.

    Examples
    --------
    >>> pr = ExponentialPrior()                     # scale=1.0
    >>> pr = ExponentialPrior( scale=5.0 )          # scale=5

    Attributes
    ----------
    scale : float
        scale of the exponential
    zeroFraction : float
        fraction of unit that map to zero in domain and vice versa.

    Hidden Attributes
    -----------------
    _shift : float
        shift of the exponential part to make room for zeros.
    _uval : float
        random number within [0,zeroFraction].

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain

    lowLimit and highLimit cannot be used in this implementation.

    Author: Do Kester.
    """

    MAXVAL = 40

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, prior=None ):
        """
        Constructor.

        Parameters:
        scale : float
            of the exponential
        prior : ExponentialPrior
            prior to copy (with new scale if applicable)

        """
        self.scale = scale
        self.zeroFraction = 0.0
        self._uval = 0.0

        super( ).__init__( domain=[0,math.inf], prior=prior )

        if prior is not None :
            self.zeroFraction = prior.zeroFraction
            self.lowLimit = prior.lowLimit
            self._uval = prior._uval

    def copy( self ):
        """ Copy the prior """
        return ExponentialPrior( scale=self.scale, prior=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: zeroFraction

        Raises
        ------
        ValueError when zeroFraction is not in [0,1]

        """
        if name == "zeroFraction" :
            if not ( 0 <= value < 1 ) :
                raise ValueError( "Fraction of zeroes must be between [0,1]" )
            setatt( self, name, float( value ) )
            setatt( self, "_shift", 1.0 - value )
            random.seed( 34567 )
            setatt( self, "_rng", random )
        elif name == "_uval" :
            setatt( self, name, float( value ) )
        else :
            super( ExponentialPrior, self ).__setattr__( name, value )

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Exponential distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        uv = 1 - uval
        if uv == 0 : return self.MAXVAL * self.scale
        if ( uv > self._shift ) :
            self._uval = self._rng.random() * self.zeroFraction     # arbitrary
            return 0
        else :
            return -math.log( uv / self._shift ) * self.scale

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Exponential distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return ( self._uval if ( dval == 0 )
                 else 1 - math.exp( -dval / self.scale ) * self._shift )


    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if self.isOutOfLimits( x ) : return 0

        return math.exp( -x / self.scale ) * self._shift / self.scale

    def logResult( self, x ):
        """
        Return a the log of the result of the distribution function to p.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if self.isOutOfLimits( x ) : return -math.inf

        return math.log( self._shift / self.scale ) - x / self.scale

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        p : float
            the value

        """
        return math.nan if self.isOutOfLimits( p ) else -1.0 / self.scale

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "Exponential prior with scale = %.2f"%self.scale +
            ( (" and zero fraction = %2f"%( self.zeroFraction ) ) if self._shift < 1 else "") )

#      * End of ExponentialPrior





import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools

from .ClassicProblem import ClassicProblem
from .Dynamic import Dynamic
from .Modifiable import Modifiable

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2021 Do Kester

class EvidenceProblem( ClassicProblem ):
    """
    An EvidenceProblem is a ClassicProblem containing a Dynamic and/or Modifiable
    model, where the (Gauss-approximated) evidence is used as likelihood

    Problems can be solved by NestedSampler, with appropriate Engines and
    ErrorDistributions.

    The result of the function for certain x and p is given by
    problem.result( x, p )
    The parameters, p, are to be optimized while the x provide additional
    information.

    Attributes from Problem
    -----------------------
    model, xdata, ydata, weights

    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None, copy=None ):
        """
        Constructor for classic problems.

        Parameters
        ----------
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with ydata
        copy : Problem
            to be copied

        """
        if not ( model.isDynamic() or model.isModifiable() ) :
            raise ValueError( "Model needs to be dynamic or modifiable" )

        super( ).__init__( model=model, xdata=xdata, ydata=ydata, weights=weights,
                           copy=copy )


    def copy( self ):
        """
        Copy.

        The copy points to the same instance of model.
        """
        return EvidenceProblem( model=self.model, xdata=self.xdata, ydata=self.ydata,
                weights=self.weights, copy=self )


    def myEngines( self ) :
        """
        Return a default list of preferred engines
        """
        engs = []

        if self.model.isDynamic() :
            engs += ["birth", "death"]

        if self.model.isModifiable() :
            engs += ["struct"]

        return engs

    def myDistribution( self ) :
        """
        Return the default preferred ModelDistribution: "model"
        """
        return "model"

    #  *****TOSTRING***********************************************************
    def baseName( self ):
        """ Returns a string representation of the model.  """
        return "EvidenceProblem of " + self.model.__str__()






import numpy as numpy
from astropy import units
import math
import warnings
from . import Tools
from .Formatter import formatter as fmt
from .Dynamic import Dynamic
from .Engine import Engine
from .ErrorsInXandYProblem import ErrorsInXandYProblem

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester

class StartEngine( Engine ):
    """
    StartEngine generates a random trial sample.

    It is used to initialize the set of trial samples.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.
        Parameters
        ----------
        copy : StartEngine
            engine to be copied

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return StartEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StartEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood ):
        """
        Execute the engine by a random selection of the parameters.

        Parameters
        ----------
        walker : Sample
            sample to diffuse
        lowLhood : float
            lower limit in logLikelihood

        Returns
        -------
        int : the number of successfull moves

        """
        walker = self.walkers[kw]
        problem = walker.problem
        model = problem.model
        fitIndex = walker.fitIndex
        allp = walker.allpars.copy()

        npar = len( allp )
        onp = problem.npars

        maxtrials = self.maxtrials
        if hasattr( self.errdis, "constrain" ) and callable( self.errdis.constrain ) :
            maxtrials *= 100

        ktry = 0
        while True :

            if model.isDynamic() :
                np0 = model.npars
                off = 0

                ## find the model in the chain that is actually dynamic
                while model is not None and not isinstance( model, Dynamic ) :
                    off += model.npbase
                    model = model._next

                npbase = model.npbase
                ## Grow the dynamic model a number of times according to growPrior
                while ( model.ncomp < model.growPrior.unit2Domain( self.rng.rand() ) and
                        model.grow( offset=off, rng=self.rng ) ) :
                    np1 = model.npars
                    dnp = np1 - np0
                    fitIndex = model.alterFitindex( fitIndex, npbase, dnp, off )
                    npbase = model.npbase
                    np0 = np1

                allp = numpy.zeros( npar + np0 - onp, dtype=float )

            uval = self.rng.rand( len( fitIndex ) )
            allp[fitIndex] = self.unit2Domain( problem, uval, kpar=fitIndex )

            logL = self.errdis.logLikelihood( problem, allp )

            if numpy.isfinite( logL ) :
                break
            elif ktry > ( maxtrials + walker.id ) :
                raise RuntimeError( "Cannot find valid starting solutions at walker %d" % walker.id )
            else :
                ktry += 1

        self.setWalker( walker.id, problem, allp, logL, fitIndex=fitIndex )

        wlkr = self.walkers[walker.id]
#        print( walker.id, wlkr.id, wlkr.problem.model.npars, len( wlkr.allpars ), fmt( wlkr.logL ) )

        nhyp = self.errdis.nphypar 
        
        nuisance = len( problem.xdata ) if isinstance( problem, ErrorsInXandYProblem ) else 0

        wlkr.check( nhyp=nhyp, nuisance=nuisance )

        return len( fitIndex )






import numpy as numpy
from astropy import units
import math
from . import Tools
from . import NeuralNetUtilities
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt
from .Formatter import fma

from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2020 - 2022 Do Kester


class SoftMaxModel( NonLinearModel ):
    """
    Softmax Model is a Logistic model if the number of outputs is 1.
    Otherwise it is generalization of the LogisticModel over multiple outputs

                       exp( sum_k( x_k * p_kn ) + q_n ) )
        f_n( x:p ) = -------------------------------------------
                     sum_i( exp( sum_k( x_k * p_ki ) + q_i ) ) )


        0       0       0       0       0       0   I inputs
        |\     /|\     /|\     /|\     /|\     /|

          all inputs connect to all outputs        I*N connecting parameters

           \|/     \|/     \|/     \|/     \|/      N offset parameters (if offset)
            0       0       0       0       0       N outputs



    The parameters (p) are initialized at 1.0, except the offset (q).
    They are initialized at 0.0.



    Attributes
    ----------
    offset : bool
        True : the outputs have offsets
    ndout : int
        number of output categories
    in2out : int
        ndim * ndout
    normed : bool
        the results are normalized (def:True)

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """

    def __init__( self, ndim=1, ndout=1, copy=None, offset=False, normed=True, **kwargs ):
        """
        Logistic response model.

        Number of parameters is npars (see offset)

        Parameters
        ----------
        ndim : int
            number of inputs
        ndout : int
            number of classifications
        offset : bool
            False : no offsets                  npars = ndim * ndout
            True  : each output has one offset: npars = ndim * ndout + ndout
        normed : bool
            True : output is normalized
            False : not
        copy : SoftMaxModel
            to be copied

        """
        in2out = ndim * ndout
        param = [1.0] * in2out
        np = in2out

        if offset :
            param += [0.0] * ndout
            np += ndout

        if offset :
            raster = NeuralNetUtilities.ConnectWithBias( ndim=ndim, ndout=ndout )
        else :
            raster = NeuralNetUtilities.Connect( ndim=ndim, ndout=ndout )

        if normed :
            if ndout == 1 :
                trans = NeuralNetUtilities.Logistic()
            else :
                trans = NeuralNetUtilities.Softmax( ndout=ndout )
        else : 
            trans = NeuralNetUtilities.Identity()

        setatt( self, "raster", raster )
        setatt( self, "trans", trans )

        setatt( self, "offset", offset )
        setatt( self, "in2out", in2out )
        setatt( self, "ndout", ndout )
        setatt( self, "normed", normed )

        super( ).__init__( np, ndim=ndim, copy=copy, params=param, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return SoftMaxModel( ndim=self.ndim, ndout=self.ndout, normed=self.normed,
                                   offset=self.offset, copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function: F_n( x_k ) as array of
        shape [nx,ndout], where nx number of data points and ndout is the number of
        outputs.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        res = self.trans.result( self.raster.result( xdata, params ), params )

        return res if self.ndout > 1 else res[:,0]

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value as a list (size N) of arrays
        of shape (K,P). N is #outputs; K is #datapoints; P is #parameters.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        dfdr = self.trans.derivative( self.raster.result( xdata, params ), params )
        drdp = self.raster.partial( xdata, params )

        part = self.trans.pipe( dfdr, drdp )

        return part[0,:,:] if self.ndout == 1 else list( part ) 

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df_i/dx_n of each output f_i to the data x_n
        at each xdata (=x).
        It is returned as an array of shape (N,I) of an array of length K.
        N is #outputs; I is #inputs (ndim); K is #datapoints.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        dfdr = self.trans.derivative( self.raster.result( xdata, params ), params )
        drdx = self.raster.derivative( xdata, params )

        der = self.trans.pipe( dfdr, drdx )

        if self.ndout == 1 :
            return der[0,:,0] if self.ndim == 1 else der[0,:,:] 
        else :
            return der[:,:,0].T if self.ndim == 1 else list( der )


    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "SoftMax: f_n( x:p ) = exp( sum_i( x_i * p_in ) + q_n)/ sum_n )" )


    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        xu = 1 if k > self.ndim * self.ndout else ( self.xUnit if self.ndim == 1 else self.xUnit[k % self.ndim] )
        return units.Unit( 1 ) / xu




import numpy as numpy
from . import Tools
import math

from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class MonteCarlo( object ):
    """
    Helper class to calculate the confidence region of a fitted model.

    MonteCarlo for models.

    The MonteCarlo class is to be used in conjunction with Model classes.

    Author:      Do Kester

    Attributes
    ----------
    xdata : array_like
        array of independent input values
    model : Model
        the model function to be fitted
    mcycles : int
        Sets number of cycles in the MonteCarlo procedure to estimate
        error bars. Default = 25

    Hidden Attributes
    -----------------
    _eigenvectors : array_like (read only)
        from eigenvalue decomposition of covariance matrix
    _eigenvalues : array_like (read only)
        from eigenvalue decomposition of covariance matrix
    _random : random
        random number generator

    """

    #  *****CONSTRUCTORS********************************************************
    def __init__( self, xdata, model, covariance, index=None, seed=12345, mcycles=25 ):
        """
        Create a new MonteCarlo, providing inputs and model.

        A MonteCarlo object is defined by its model and the input vector (the
        independent variable). When a fit to another model and/or another
        input vector is needed a new object should be created.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
        covariance : matrix
            the covariance matrix of the problem. Default from the Model.
        index : list of int
            indices of parameters to fit
        seed : int
            seed for random number generator
        mcycles : int
            number of cycles in the MonteCarlo procedure to estimate error bars.

        Raises
        ------
        ValueError when model and input have different dimensions

        """

        self.xdata = xdata
        self.nxdata = Tools.length( xdata )
        self.model = model
        self.index = index
        self.mcycles = mcycles
        self._random = numpy.random
        self._random.seed( seed )
        self.decompose( covariance )

    def decompose( self, covariance ):
        val, vec = numpy.linalg.eigh( covariance )
        self._eigenvectors = vec
        self._eigenvalues = numpy.sqrt( val )

    #  *****MONTE CARLO ERROR***************************************************
    def getError( self, xdata=None ):
        """
        Calculates 1 &sigma;-confidence regions on the model given some inputs.

        From the full covariance matrix ( = inverse of the Hessian ) random
        samples are drawn, which are added to the parameters. With this new
        set of parameters the model is calculated. This procedure is done
        by default, 25 times.
        The standard deviation of the models is returned as the error bar.

        Parameters
        ----------
        xdata ; array_like
            input data over which to calculate the error bars. default provided xdata

        """
        if xdata is None :
            xdata = self.xdata
        ninp = Tools.length( xdata )
        sm1 = numpy.zeros( ninp )
        sm2 = numpy.zeros( ninp )

        for k in range( self.mcycles ) :
            model = self.randomVariant( xdata )
            sm1 += model
            sm2 += numpy.square( model )
        sm1 /= self.mcycles
        sm2 /= self.mcycles
        return numpy.sqrt( sm2  - sm1 * sm1 )

    def randomVariant( self, xdata ):
        """
        Return a random variant of the model result.
        Taking into account the stdev of the parameters and their covariance.

        Parameters
        ----------
        xdata : array_like
            input data at these indpendent points

        """
        nfit = self.model.npchain if self.index is None else len( self.index )
        err = self._random.standard_normal( nfit )
        err = numpy.inner( self._eigenvectors, self._eigenvalues * err )
        par = self.model.parameters.copy()
        if self.index is None :
            par += err
        else :
            par[self.index] += err

        return self.model.result( xdata, par )

    #  *************************************************************************
    def __str__( self ):
        """ Return name of the class.  """
        return "MonteCarlo with seed %d and %d cycles" % (self._random.seed, self.mcycles)







import numpy as numpy
import math

from .BaseFitter import BaseFitter

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class QRFitter( BaseFitter ):
    """
    Fitter for linear models, using QR decomposition.
    The QRFitter class is to be used in conjunction with Model classes, linear
    in its parameters.

    For Linear models the matrix equation

        H * p = &beta;

    is solved for p. H is the Hessian matrix ( D * w * D^T )
    and &beta; is the inproduct of the data with the D, design matrix.

        &beta; = y * w * D^T

    The QRFitter class use QR decomposition which effectively is an inversion
    of the hessian matrix such that

        p = &beta; * inverse( H )

    It can be more efficient if
    similar ydata needs to be fitter to the same model and xdata.
    In that case it uses the same decomposition for all fits.

    Examples
    --------
    # assume x and y are numpy.asarray data arrays:
    >>> x = numpy.asarray.range( 100 )
    >>> poly = PolynomialModel( 1 )                             # line
    >>> fitter = QRFitter( x, poly )
    >>> for k in range( 1, 4 ) :
    >>>     y = numpy.arange( 100 ) // k                        # digitization noise
    >>>     param = fitter.fit( y )                             # use same QR decomposition
    >>>     stdev = fitter.stdevs                               # stdevs on the parameters
    >>>     print( k, param )
    >>>     print( " ", stdev )

    Category:    Mathematics/Fitting

    Attributes
    ----------
    needsNewDecomposition : bool
        True when starting. Thereafter False,
            i.e. the previous QR-decomposition is used, unless weights are used.
            Only for linear fitters, setting it to false might improve efficiency.
            It only works properly when the model, the x-data etc are exactly the
            same in the previous run.
            Whenever weights are used, it always needs a new decomposition.

    qrmat : matrix
        matrix formed by q * inverse( r ), where q,r is the QR decomposition
        of the design matrix.
        qrmat is to be multiplied with the data vector to get the solution.

    """

    def __init__( self, xdata, model, map=False, keep=None ):
        """
        Create a new Fitter, providing xdatas and model.

        A Fitter class is defined by its model and the input vector (the
        independent variable). When a fit to another model and/or another
        input vector is needed a new object should be created.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
        map : bool (False)
            When true, the xdata should be interpreted as a map.
            The fitting is done on the pixel indices of the map,
            using ImageAssistant
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep will be used by the Fitter as long as the Fitter exists.
            See also `fit( ..., keep=dict )`

        """
        super( QRFitter, self ).__init__( xdata, model, map=map, keep=keep )

        self.needsNewDecomposition = True
        self.qrmat = None

    def fit( self, ydata, weights=None, keep=None ):
        """
        Return model parameters fitted to the data, including weights.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data ( = 1.0 / sigma^2 )
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.
        Raises
        ------
        ValueError when ydata or weights contain a NaN

        """
        fi, ydata, weights = self.fitprolog( ydata, weights=weights, keep=keep )

        if self.model.isNullModel() :
            self.chiSquared( ydata, weights )
            return numpy.asarray( 0 )

        ydatacopy = ydata.copy( )
        # subtract influence of fixed parameters on the data
        if fi is not None :
            fxpar = numpy.copy( self.model.parameters )
            fxpar[fi] = 0.0
            ydatacopy = numpy.subtract( ydatacopy, self.model.result( self.xdata, fxpar) )

        wgts = ( numpy.ones_like( ydata, dtype=float ) if weights is None else
                 numpy.sqrt( weights ) )

        if hasattr( self, "normdata" ) :
            normweight = math.sqrt( self.normweight )
            ydatacopy = numpy.append( ydatacopy, self.normdata )
            wgts = numpy.append( wgts, normweight )

        ydatacopy = ydatacopy * wgts

        if self.needsNewDecomposition or weights is not None:
            design = ( self.getDesign( index=fi ).transpose() * wgts ).transpose()

            # The QR decomposition in numpy has a different interface from that in scipy.
            # Here we use numpy. See:
            #   https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.qr.html

            q, r = numpy.linalg.qr( design )
            self.qrmat = numpy.dot( numpy.linalg.inv( r ), q.transpose() )
            self.needsNewDecomposition = False

        params = numpy.dot( self.qrmat, ydatacopy )

        params = self.insertParameters( params, index=fi )
        self.model.parameters = params
        self.chiSquared( ydata, weights=weights )

        return params

#  *************************************************************************
    def __str__( self ):
        """ Return the name of the fitter. """
        return "QRFitter"


